/*  findnull.cpp

    Find likely nulls.  This C++ program is really pretty compatible to C.
    
    findnull.cpp <file 1> <file 2> ... <file n>
*/

#include <stdio.h>
#include <math.h>

// A function to find and print nulls.  It returns 0 if no error, 1 if error
int findNulls (char *ipcFileName);

int main(int argc, char *argv[]) {
    int iFileNumber;
    int iResult = 0;

    for (iFileNumber = 1; iFileNumber < argc; ++iFileNumber) {
        fprintf(stderr,"%s\n",argv[iFileNumber]);
        iResult += findNulls(argv[iFileNumber]);
    }

    return iResult;
}

// A function to find and print nulls.  This uses a bit of memory, and it could
// be more efficient, but that would take time I don't have at the moment.
// On the other hand, it uses less memory than 90% of the processes currently
// running on my machine, so...
int findNulls (char *ipcFileName) {
    FILE *pfStream;

    unsigned int uiTotalCharacters = 0;
    unsigned int uiSingle[0x100];
    unsigned int uiPair[0x100][0x100];
    double dScore[0x100];

    int c;
    int d;
    int e;

    double dNormal;
    double dBefore;
    double dAfter;

    // Open the file
    pfStream = fopen(ipcFileName, "rb");

    if (pfStream == NULL || ferror(pfStream)) {
        fprintf(stderr,"** Error opening file %s\n", ipcFileName);
        return 1;
    }

    // Initialize the arrays
    for (c = 0; c < 0x100; ++c) {
        uiSingle[c] = 0.0;
        for (d = 0; d < 0x100; ++d) {
            uiPair[c][d] = 0.0;
        }
    }

    // Count the characters and pairs
    c = fgetc(pfStream);
    ++uiTotalCharacters;
    ++uiSingle[c];

    d = fgetc(pfStream);

    while (! feof(pfStream) && ! ferror(pfStream)) {
        // Ignore whitespace
        if (d > ' ' && c > ' ') {
            ++uiTotalCharacters;
            ++uiSingle[d];
            ++uiPair[c][d];
        }

        c = d;
        d = fgetc(pfStream);
    }

    fclose(pfStream);

    // Print out a NULL score for each character.  The higher this number,
    // the more likely that the character is a NULL
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0) {
            dScore[c] = 0.0;
    
            for (d = 0; d < 0x100; ++d) {
                // How commonly does d occur?
                dNormal = (double)(uiSingle[d]) / (double)(uiTotalCharacters);
                
                // How commonly does d occur before c?
                dBefore = (double)(uiPair[d][c]) / (double)(uiSingle[c]);
                
                // How commonly does d occur after c?
                dAfter = (double)(uiPair[c][d]) / (double)(uiSingle[c]);

                // What's the score?  The total number will be high only if
                // there is a high correlation between the relative frequencies
                // of the bytes throughout the text and the bytes as they
                // occur before and after our test byte
                if (dNormal < dBefore) {
                    dScore[c] += dNormal;
                } else {
                    dScore[c] += dBefore;
                }
            
                if (dNormal < dAfter) {
                    dScore[c] += dNormal;
                } else {
                    dScore[c] += dAfter;
                }
            }
        }
    }

    fprintf(stdout, "    C   (Hex )   Score      Assessment\n");
    fprintf(stdout, "    -   ------   --------   -------------------\n");

    // Over 1.9, probably NULL
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0 && dScore[c] >= 1.9) {
            fprintf(stdout, "    %c   (0x%2x)   %f   Probably null\n", c, c, dScore[c]);
        }
    }

    // Over 1.5, maybe NULL
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0 && dScore[c] >= 1.5 && dScore[c] < 1.9) {
            fprintf(stdout, "    %c   (0x%2x)   %f   Maybe null\n", c, c, dScore[c]);
        }
    }

    // Over 1.3, borderline
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0 && dScore[c] >= 1.3 && dScore[c] < 1.5) {
            fprintf(stdout, "    %c   (0x%2x)   %f   Questionable null\n", c, c, dScore[c]);
        }
    }

    // Over 0.8, probably not null
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0 && dScore[c] >= 0.8 && dScore[c] < 1.3) {
            fprintf(stdout, "    %c   (0x%2x)   %f   Probably not null\n", c, c, dScore[c]);
        }
    }

    // Under 0.8, definitely not null
    for (c = 0; c < 0x100; ++c) {
        if (uiSingle[c] != 0 && dScore[c] < 0.8) {
            fprintf(stdout, "    %c   (0x%2x)   %f   definitely not null\n", c, c, dScore[c]);
        }
    }

    return 0;
}

